
/*
=================================================================

  err_skel.c: PCYACC LALR(1) parser driver 
	      with syntax error diagnostics 
  by Daming Liu
  

  PCYACC (R) is a software product of ABRAXAS SOFTWARE INC.
  Copyright(C) 1986-1996 by ABRAXAS SOFTWARE INC.

=================================================================
*/


/*
--------------------------------------------
  the following macros are supplied 
  for compatibility with UNIX yacc
--------------------------------------------
*/  
  
#define  yyerrok         error_flag = 0           
#define  yyclearin       if (yytoken == 0)  return  1; else  yytoken = -1
#define  YYACCEPT        return (0)
#define  YYABORT         return (1)

#define  YYFLAG          -1000         /* this symbol is required by PCYACC */
#define  YYERROR         goto yyerrlab

#ifndef  YYMAXDEPTH
#define  YYMAXDEPTH      100                             /* depth of parser */
#endif


/*
--------------------------------------------
  globals, externals
--------------------------------------------
*/

extern char *yydisplay(int);
extern void yyerror(char *, char *);
		     
YYSTYPE  yyv[YYMAXDEPTH];                                    /* token stack */

int yytoken        = -1;                                 /* lookahead token */
int error_count    =  0;                                 /* count of errors */
int error_flag     =  0;                                      /* error flag */



int yyparse( void )                                        /* syntax parser */
{
    int yyj;
    int yym;
    int yyn;
    int yystate;
    
    int yys[YYMAXDEPTH];

    int *yyps;
    int const *yyxi;

    YYSTYPE *yypv;
    YYSTYPE *yypvt;
    
#ifdef YYDEBUG
    printf("state 0\n");
#endif

    yytoken     = -1;
    error_count =  0;
    error_flag  =  0;
    yystate     =  0;
    
    yypv  =  yyv - 1;
    yyps  =  yys - 1;
    

yystack:                            /* put a state and value onto the stack */                        

    if ( ++yyps > yys + YYMAXDEPTH)                     /* fatal condiction */
    {
	printf("payacc stack overflow", 0);
	return (1);
    }  

    *yyps = yystate;
    ++yypv;
    *yypv = yyval;
    
    
yynewstate:
  
    yyn = yypact[yystate];
    

    if ( yyn <= YYFLAG )           /* simple state with only default action */
    { 
	goto yydefault;
    }
    
    if ( yytoken < 0)                                /* no lookahead token? */
    {
	if ( (yytoken = yylex()) < 0) 
	{
	    yytoken = 0;
	}
    }     
    
    yyn += yytoken;
    
    if ( yyn < 0 || yyn >= YYLAST )               /* no entry within table? */
    {
	goto yydefault;
    }
    
    if ( yychk[ yyn = yyact[yyn] ] == yytoken )             /* valid shift? */
    {
    
#ifdef YYDEBUG    
	printf("shift on [%s] to state %d\n", yydisplay(yytoken), yyn);
#endif

	yytoken = -1;                            /* lookahead token is read */
	yyval = yylval;
	yystate = yyn;
 
	if ( error_flag > 0 )                      /* still in error state? */
	{
	    --error_flag;
	}

	goto yystack;
    }
     
     

yydefault:                         /* simple state with only default action */

    if ((yyn = yydef[yystate]) == -2)
    {
	if ( yytoken < 0)                            /* no lookahead token? */
	{
	    if ( (yytoken = yylex()) < 0) 
	    {
		 yytoken = 0;
	    }
	}

	/* 
	--------------------------------
	 look through exception table 
	--------------------------------
	*/
	 
	for ( yyxi = yyexca; *yyxi != -1 || yyxi[1] != yystate; yyxi +=2)
	    ;                                                       /* void */
	      
	while ( *(yyxi += 2) >= 0)
	{
	    if ( *yyxi == yytoken )
	    {
		 break;
	    }
	}
 
	if ( (yyn = yyxi[1]) < 0 )
	{
	 
#ifdef YYDEBUG
	    printf("accept\n");
#endif           
	 
	    return 0;                                             /* accept */
	}
    }
	
    if ( yyn == 0)                                                /* error? */
    {
	switch( error_flag)
	{
	    case 0:                                      /* brand new error */
      
		      yyerror( "syntax error", yydisplay(yytoken) );

		      if ((yyn = yypact[yystate]) > YYFLAG && yyn < YYLAST)
		      {
			   register int x;
			 
			   for ( x = (yyn > 0) ? yyn : 0;   x < YYLAST;  ++x ) 
			   {
				if ( yychk[yyact[x]] == x-yyn       && 
				     x-yyn           != YYERRCODE      )
				{
				     yyerror ( 0, yydisplay(x - yyn) );
				}      
			   }
			
			   yyerror(0, 0);
		      }
	 
		      yyerror(0, 0);

yyerrlab:
		      ++error_count;
		   
	    case 1:
		   
		   
	    case 2:              /* incompletely recovered error, try again */
	      
		      error_flag = 3;         /* find a state where "error" 
						 is a legal shift action    */
		    
		      while( yyps >= yys )
		      {
			  yyn = yypact[*yyps] + YYERRCODE;
			
			  if( yyn               >= 0          &&
			      yyn               <  YYLAST     &&
			      yychk[yyact[yyn]] == YYERRCODE     )
			  {
			      yystate = yyact[yyn];     /* simulate a shift
							   of error         */
			     
#ifdef YYDEBUG
			      printf("shift on [error] to state %d\n",
				      yystate);
#endif           

			      goto yystack;
			     
			  }
			
			  yyn = yypact[*yyps];
			
			  --yyps;                   /* the current yyps has
						       no shift on "error"  */ 
   
			  --yypv;                       /* so pop the stack */
			
#ifdef YYDEBUG
			  printf("error: pop state %d, uncovering state %d\n", 
					      yyps[1],              yyps[0]);
#endif           
		      }
		   
		    
		      /*
		      --------------------------------------------------- 
		      there is no state on the stack with an error shift, 
		      so abort 
		      --------------------------------------------------- 
		      */        
						   
yyabort:

#ifdef YYDEBUG
		      printf("abort\n");
#endif           
		   
		      return 1;
		   
	     case 3:                    /* no shift yet, clobber input char */

#ifdef YYDEBUG
		      printf("error: discard [%s]\n", yydisplay(yytoken)); 
#endif                   
	 
		      if ( yytoken == 0 )                           /* EOF? */
		      {
			       goto yyabort;           /* don't discard EOF,
							  quit instead      */
		      }
	 
		      yytoken = -1;          /* try again in the same state */
		      goto yynewstate;
	}  
    }
     
	     
    /*
    ------------------------------- 
      reduction by production yyn 
    -------------------------------
    */


    yyps -= yyr2[yyn];
    yypvt = yypv;
    yypv -= yyr2[yyn];
    yyval = yypv[1];
    yym   = yyn;
     
    /*
    -----------------------------------------  
      consult goto table to find next state  
    ----------------------------------------- 
    */
     
    yyn = yyr1[yyn];
    yyj = yypgo[yyn] + *yyps + 1;
     
    if ( yyj >= YYLAST || yychk[yystate = yyact[yyj]] != -yyn )
    {
	 yystate = yyact[yypgo[yyn]];
    }
     
#ifdef YYDEBUG
    printf("reduce w/ rule %d, uncover state %d, go to state %d\n", 
			  yym,            *yyps,          yystate); 
#endif           
  
    switch (yym)
    {
	$A
    }
     
    goto yystack ;                             /* stack new state and value */
} 
     
   
