/*
==========================================================================

  JAVA.Y: PCYACC grammar description for JAVA parser
  version 1.0
 
  by Daming Liu 
  

  PCYACC (R) is a software product of ABRAXAS SOFTWARE INC.
  Copyright (C) 1986-1996 by ABRAXAS SOFTWARE INC.

  Reference:    The Java(TM) Language Specification 
                Version 1.0 Beta           
                Oct. 30, 1995
                Sun Microsystems Computer Corporation
  
  				Java!
  				by Tim Ritchey
  				
  				Hooked on JAVA
  				by Arthur van Hoff, Sami Shaio, and Orca Starbuck
  
  				60 Minute Guide to Java
  			 	by Ed Tittel & Mark Gaither

===========================================================================
*/

%{
#include <stdio.h>

extern void  yyerror(char *, char *);
extern char *yydisplay(int);
extern int   yylex(void);
%}

%union {
  int   i;
  float r;
  char *s;
}

/*
================================================= 

               TOKEN DECLARATION

================================================= 
*/
/*
================================================= 
               RESERVED KEYWORDS
================================================= 
*/

%token	ABSTRACT
%token	BOOLEAN
%token	BREAK
%token	BYTE
%token	BYVALUE		/* designates reserved, but not implemented yet. */
%token	CASE
%token	CAST		/* designates reserved, but not implemented yet. */
%token	CATCH
%token	CHAR
%token	CLASS
%token	CONST		/* designates reserved, but not implemented yet. */
%token	CONTINUE
%token	DEFAULT
%token	DO
%token	DOUBLE
%token	ELSE
%token	EXTENDS
%token	FALSE
%token	FINAL
%token	FINALLY
%token	FLOAT
%token	FOR
%token	FUTURE		/* designates reserved, but not implemented yet. */
%token	GENERIC		/* designates reserved, but not implemented yet. */
%token	GOTO   		/* designates reserved, but not implemented yet. */
%token	IF
%token	IMPLEMENTS
%token	IMPORT
%token	INNER		/* designates reserved, but not implemented yet. */
%token	INSTANCEOF
%token	INT
%token	INTERFACE
%token	LONG
%token	NATIVE
%token	NEW
%token	__NULL__
%token	OPERATOR	/* designates reserved, but not implemented yet. */
%token	OUTER		/* designates reserved, but not implemented yet. */
%token	PACKAGE
%token	PRIVATE
%token	PROTECTED
%token	PUBLIC
%token	REST		/* designates reserved, but not implemented yet. */
%token	RETURN
%token	SHORT
%token	STATIC
%token	SUPER
%token	SWITCH
%token	SYNCHRONIZED
%token	THIS
%token	THREADSAFE
%token	THROW
%token	THROWS		
%token	TRANSIENT
%token	TRUE
%token	TRY
%token	VOID
%token	VOLATILE		
%token	WHILE


/*
================================================= 
                   CONSTANTS
================================================= 
*/
%token	__INTEGER_CONSTANT 
%token	__FLOAT_CONSTANT 
%token	__CHARACTER_CONSTANT 
%token	__STRING


/*
================================================= 
                   IDENTIFIER
================================================= 
*/

%token	__IDENTIFIER 



/* 
================================================= 
       COMBINED OPERATORS AND SEPARATORS
=================================================
*/ 
%token	ADDADD							/* ++   */
%token	SUBSUB							/* --   */
%token	EQU								/* ==   */
%token	LEQ								/* <=   */
%token	GEQ								/* >=   */
%token	NEQ								/* !=   */
%token	SHL								/* <<   */
%token	SHR								/* >>   */   
%token	ZFSHR							/* >>>  */
%token	ADDEQ							/* +=   */	
%token	SUBEQ							/* -=   */	
%token	MULEQ							/* *=   */	
%token	DIVEQ							/* /=   */	
%token	ANDEQ							/* &=   */	
%token	IOREQ							/* |=   */
%token	XOREQ							/* ^=   */	
%token	MODEQ							/* %=   */
%token	SHLEQ							/* <<=  */	
%token	SHREQ							/* >>=  */
%token	ZFSHRE							/* >>>= */	
%token	OROR							/* ||   */
%token	ANDAND							/* &&   */


/* 
================================================= 
                OPERATOR PRECEDENCE
=================================================
*/ 

%nonassoc Shift
%nonassoc error  
%nonassoc __IDENTIFIER  
%nonassoc ELSE  CATCH  FINALLY 

/* 
 * comma operator 
*/
%left	','


/* 
 * assignment operators 
*/
%right	'='
	ADDEQ							/* +=   */	
	SUBEQ							/* -=   */	
	MULEQ							/* *=   */	
	DIVEQ							/* /=   */	
	ANDEQ							/* &=   */	
	IOREQ							/* |=   */
	XOREQ							/* ^=   */	
	MODEQ							/* %=   */
	SHLEQ							/* <<=  */	
	SHREQ							/* >>=  */
	ZFSHRE							/* >>>= */	


/* 
 * binary operators 
*/
%right	'?'	':'
%left	OROR
%left	ANDAND 
%left   '|'
%left  	'^'
%left	'&'
%left	EQU  NEQ 
%left	LEQ	 GEQ  '<'  '>'
%left	SHL	 SHR  ZFSHR
%left	'+'  '-'
%left	'*'  '/'  '%'


/* 
 * unary operators 
*/
%right	ADDADD	SUBSUB	'~'	'!'	INSTANCEOF	


/* 
 * special operators 
*/
%left	'('	'['	'.'	')'



/* 
================================================= 
                START SYMBOL
=================================================
*/ 
%start CompilationUnit



%%



CompilationUnit
  : PackageStatement  ImportStatementList  TypeDeclarationList
  |                   ImportStatementList  TypeDeclarationList  
  | PackageStatement    	               TypeDeclarationList
  |                                        TypeDeclarationList  
  | error			              /* last ditch error recovery */
  ;

PackageStatement
  : PACKAGE  Name ';'
  ;
  
  
ImportStatementList
  :                     ImportStatement
  | ImportStatementList ImportStatement
  ; 
  
  
TypeDeclarationList
  :                     TypeDeclaration
  | TypeDeclarationList TypeDeclaration
  ;


ImportStatement
  : IMPORT  Name '.' '*' ';'
  | IMPORT  Name         ';'
  ;
  
TypeDeclaration
  : ClassDeclaration
  | InterfaceDeclaration
  | ';'
  ;  

ClassDeclaration
  : ModifierList       CLASS                   __IDENTIFIER  
    ExtendsStatement1  ImplementsStatement '{' FieldDeclarationList '}'
    
  | ModifierList       CLASS                   __IDENTIFIER  
    ExtendsStatement1                      '{' FieldDeclarationList '}'
    
  | ModifierList       CLASS                   __IDENTIFIER                    
                       ImplementsStatement '{' FieldDeclarationList '}'
                      
  | ModifierList       CLASS                   __IDENTIFIER  
                                           '{' FieldDeclarationList '}'

  | ModifierList       CLASS                   __IDENTIFIER  
    ExtendsStatement1  ImplementsStatement '{'                      '}'
    
  | ModifierList       CLASS                   __IDENTIFIER  
    ExtendsStatement1                      '{'                      '}'
    
  | ModifierList       CLASS                   __IDENTIFIER                    
                       ImplementsStatement '{'                      '}'
                      
  | ModifierList       CLASS                   __IDENTIFIER  
                                           '{'                      '}'

  |                    CLASS                   __IDENTIFIER  
    ExtendsStatement1  ImplementsStatement '{' FieldDeclarationList '}'
    
  |                    CLASS                   __IDENTIFIER  
    ExtendsStatement1                      '{' FieldDeclarationList '}'
    
  |                    CLASS                   __IDENTIFIER                    
                       ImplementsStatement '{' FieldDeclarationList '}'
                      
  |                    CLASS                   __IDENTIFIER  
                                           '{' FieldDeclarationList '}'

  |                    CLASS                   __IDENTIFIER  
    ExtendsStatement1  ImplementsStatement '{'                      '}'
    
  |                    CLASS                   __IDENTIFIER  
    ExtendsStatement1                      '{'                      '}'
    
  |                    CLASS                   __IDENTIFIER                    
                       ImplementsStatement '{'                      '}'
                      
  |                    CLASS                   __IDENTIFIER  
                                           '{'                      '}'
  ;         

ModifierList
  :               Modifier                   
  | ModifierList  Modifier                   
  ;
  
       
ExtendsStatement1
  : EXTENDS       Name
  ;
  
ImplementsStatement
  : IMPLEMENTS    NameList
  ;
  
NameList
  :               Name
  | NameList ';'  Name 
  ;
  
FieldDeclarationList
  :                       FieldDeclaration
  | FieldDeclarationList  FieldDeclaration 
  ;

InterfaceDeclaration
  : ModifierList            INTERFACE                   __IDENTIFIER  
    ExtendsStatement2   '{' FieldDeclarationList '}'
    
  | ModifierList            INTERFACE                   __IDENTIFIER  
                        '{' FieldDeclarationList '}'

  | ModifierList            INTERFACE                   __IDENTIFIER  
    ExtendsStatement2   '{'                      '}'
    
  | ModifierList            INTERFACE                   __IDENTIFIER  
                        '{'                      '}'  
                        
  |                         INTERFACE                   __IDENTIFIER  
    ExtendsStatement2   '{' FieldDeclarationList '}'
    
  |                         INTERFACE                   __IDENTIFIER  
                        '{' FieldDeclarationList '}'

  |                         INTERFACE                   __IDENTIFIER  
    ExtendsStatement2   '{'                      '}'
    
  |                         INTERFACE                   __IDENTIFIER  
                        '{'                      '}'
  ;   
  
ExtendsStatement2
  : EXTENDS  NameList
  ;
  

FieldDeclaration
  : MethodDeclaration
  | ConstructorDeclaration
  | VariableDeclaration
  | StaticInitializer    
  | ';'
  | error
  ;  
        
MethodDeclaration
  : ModifierList Type __IDENTIFIER '(' ParameterList ')'  BList EndMethod
  | ModifierList Type __IDENTIFIER '(' ParameterList ')'        EndMethod
  | ModifierList Name __IDENTIFIER '(' ParameterList ')'  BList EndMethod
  | ModifierList Name __IDENTIFIER '(' ParameterList ')'        EndMethod
  |              Type __IDENTIFIER '(' ParameterList ')'  BList EndMethod
  |              Type __IDENTIFIER '(' ParameterList ')'        EndMethod
  |              Name __IDENTIFIER '(' ParameterList ')'  BList EndMethod
  |              Name __IDENTIFIER '(' ParameterList ')'        EndMethod
  ;

ParameterList
  :
  |                   Parameter
  | ParameterList ',' Parameter  
  ;
  
  
BList
  :       '[' ']'
  | BList '[' ']'
  ;      
  
EndMethod
  : '{' StatementList '}'  
  | '{'               '}'  
  | ';'
  ;
  
StatementList
  :                Statement
  | StatementList  Statement
  ; 
  
ConstructorDeclaration
  : ModifierList __IDENTIFIER '(' ParameterList ')' EndMethod
  |              __IDENTIFIER '(' ParameterList ')' EndMethod
  ;
  
VariableDeclaration
  : ModifierList Type VariableDeclaratorList ';' 
  | ModifierList Name VariableDeclaratorList ';' 
  |              Type VariableDeclaratorList ';' 
  |              Name VariableDeclaratorList ';' 
  ;
  
VariableDeclaratorList
  :                            VariableDeclarator
  | VariableDeclaratorList ',' VariableDeclarator
  ;                            
  
VariableDeclarator
  : __IDENTIFIER BList
  | __IDENTIFIER BList '=' VariableInitializer
  | __IDENTIFIER 
  | __IDENTIFIER       '=' VariableInitializer
  ;
  
VariableInitializer
  :     Expression                                %prec Shift
  | '{' VariableInitializerList '}'
  ;
  
VariableInitializerList
  :
  |                             VariableInitializer
  | VariableInitializerList ',' VariableInitializer
  ;
  
     
StaticInitializer
  : STATIC  '{'  StatementList '}'
  | STATIC  '{'                '}'
  ;
  

Parameter
  : Type __IDENTIFIER BList
  | Type __IDENTIFIER 
  | Name __IDENTIFIER BList
  | Name __IDENTIFIER   
  | error
  ;
  
Statement
  : VariableDeclaration
  | Expression ';'
  | '{' StatementList '}'
  | '{'               '}'

  | IF    '(' Expression ')' Statement                     %prec Shift
  | IF    '(' Expression ')' Statement  ELSE  Statement

  | WHILE '(' Expression ')' Statement  

  | DO Statement WHILE '(' Expression ')' Statement  ';'  

  | FOR '(' Expression ';' Expression ';' Expression ')' Statement
  | FOR '(' Expression ';' Expression ';'            ')' Statement
  | FOR '(' Expression ';'            ';' Expression ')' Statement
  | FOR '(' Expression ';'            ';'            ')' Statement
  | FOR '('            ';' Expression ';' Expression ')' Statement
  | FOR '('            ';' Expression ';'            ')' Statement
  | FOR '('            ';'            ';' Expression ')' Statement
  | FOR '('            ';'            ';'            ')' Statement
  | FOR '(' VariableDeclaration  Expression ';' Expression ')' Statement
  | FOR '(' VariableDeclaration  Expression ';'            ')' Statement
  | FOR '(' VariableDeclaration             ';' Expression ')' Statement
  | FOR '(' VariableDeclaration             ';'            ')' Statement 
  
  | TRY Statement                                      %prec Shift
  | TRY Statement CatchStatementList                   %prec Shift
  | TRY Statement                    FINALLY Statement
  | TRY Statement CatchStatementList FINALLY Statement

  | SWITCH '(' Expression ')' '{'                   '}'
  | SWITCH '(' Expression ')' '{' CaseStatementList '}'

  | SYNCHRONIZED '(' Expression ')' Statement

  | RETURN Expression ';'
  | RETURN            ';'

  | THROW Expression  ';'  

  | __IDENTIFIER ':' Statement

  | BREAK  __IDENTIFIER ';'
  | BREAK               ';'

  | CONTINUE  __IDENTIFIER ';'
  | CONTINUE               ';'   
  | ';'  
  | __IDENTIFIER '[' __IDENTIFIER ']' '=' Expression ';'    
  | error
  ;

CatchStatementList
  :                    CatchStatement
  | CatchStatementList CatchStatement
  ;
  
CatchStatement
  : CATCH '('  Parameter  ')'  Statement
  ;
  
CaseStatementList
  :                    CaseStatement
  | CaseStatementList  CaseStatement
  ;

CaseStatement
  : CASE  Expression ':'
  | DEFAULT ':'
  | Statement
  ;
  
Expression
  : Expression '+' Expression  
  | Expression '-' Expression  
  | Expression '*' Expression  
  | Expression '/' Expression  
  | Expression '%' Expression  
  | Expression '^' Expression  
  | Expression '&' Expression  
  | Expression '|' Expression  
  | Expression ANDAND Expression  
  | Expression OROR   Expression  
  | Expression SHL    Expression  
  | Expression SHR    Expression  
  | Expression ZFSHR  Expression  
  | Expression '='    Expression  
  | Expression ADDEQ  Expression  
  | Expression SUBEQ  Expression 
  | Expression MULEQ  Expression  
  | Expression DIVEQ  Expression  
  | Expression MODEQ  Expression  
  | Expression XOREQ  Expression  
  | Expression ANDEQ  Expression  
  | Expression IOREQ  Expression  
  | Expression SHLEQ  Expression  
  | Expression SHREQ  Expression  
  | Expression ZFSHRE Expression  
  | Expression '<'    Expression  
  | Expression '>'    Expression  
  | Expression LEQ    Expression  
  | Expression GEQ    Expression  
  | Expression EQU    Expression  
  | Expression NEQ    Expression  
  | Expression '.'    Expression  
  | Expression ','    Expression  
  
  |  '-' Expression  
  |  '!' Expression  
  |  '~' Expression  
  |  '(' Expression ')' 
  |  '(' Type ')' Expression  
  |  '(' Name ')' Expression  

  | Expression INSTANCEOF Name
   
  | Expression '?' Expression ':' Expression 
  | Expression '[' Expression ']' 

  | ADDADD Expression  
  | SUBSUB Expression  

  | Expression ADDADD  
  | Expression SUBSUB  


  | Expression '('         ')'  
  | Expression '(' ArgList ')'  
  
  | NEW Name '(' ArgList ')'  
  | NEW Name '('         ')'  
  | NEW TypeSpecifier BExpressionList BList      %prec Shift
  | NEW Name          BExpressionList BList      %prec Shift
  | NEW TypeSpecifier BExpressionList            %prec Shift 
  | NEW Name          BExpressionList            %prec Shift
  | NEW  '(' Expression ')'  

  | TRUE
  | FALSE
  | __NULL__
  | SUPER
  | THIS                    

  | Name                                 %prec Shift
  | __INTEGER_CONSTANT
  | __FLOAT_CONSTANT
  | __STRING
  | __CHARACTER_CONSTANT   
  ;

  
BExpressionList
  :                    BExpression
  | BExpressionList    BExpression
  ;  
  
BExpression
  : '[' Expression ']'
  ;

ArgList
  : ExpressionList
  ;
  
ExpressionList
  :                    Expression                    %prec Shift
  | ExpressionList ',' Expression
  ;
  
Type
  : TypeSpecifier    BList
  | Name             BList
  | TypeSpecifier  
  ;
  
TypeSpecifier
  : BOOLEAN
  | BYTE
  | CHAR
  | SHORT
  | INT
  | FLOAT
  | LONG
  | DOUBLE
  ;
  
Modifier
  : PUBLIC
  | PRIVATE
  | PROTECTED
  | STATIC
  | FINAL
  | NATIVE
  | SYNCHRONIZED
  | ABSTRACT
  | THREADSAFE
  | TRANSIENT  
  | VOID 
  ;
  
Name
  :          __IDENTIFIER
  | Name '.' __IDENTIFIER     
  ;
  


