%{
/*
=================================================================

  lex.c: lexical analyzer for JAVA language engine
  Version 1.0
  By Daming Liu

  PCYACC (R) is a software product of ABRAXAS SOFTWARE INC.
  Copyright(C) 1986-1996 by ABRAXAS SOFTWARE INC.

=================================================================
*/

#include <stdlib.h>
#include <string.h>             /* strcpy(), getc(),  putc(), _stricmp() */
#include "java.h"              /* token  values */
#define DIM(a)	(sizeof(a)/sizeof((a)[0]))

extern unsigned char *yytext;   /* current token's text, 
                                   internal pcyacc variable */
extern FILE 	*yyin;          /* pointer to input file, defined in main.c */
extern int 	error_count;	    /* count of errors, defined in err_skel.c */

int   yylineno  = -1;			/* current line number */

%}

blank				[ \t]
letter				[a-zA-Z_$]
alphanum			[a-zA-Z_0-9$]

ODigit				[0-7]
DDigit				[0-9]
HDigit 				[0-9a-fA-F]

UnicodeChar         \\u[0-9a-fA-F]{4}
HChar				\\x[0-9a-fA-F]{2} 
OChar				\\[0-7]{3}

esc					\\([bfnrt"'?\\]|[0-7]{1,3}|x[0-9a-fA-F]+|u[0-9a-fA-F]{4})   


other				.

%x					COMMENT1
%x					COMMENT2




%%   

"++"	return ADDADD;
"--"	return SUBSUB;
"=="	return EQU;
"<="	return LEQ;
">="	return GEQ;
"!="	return NEQ;
"<<"	return SHL;
">>"	return SHR;
">>>"	return ZFSHR;
"+="	return ADDEQ;
"-="	return SUBEQ;
"*="	return MULEQ;
"/="	return DIVEQ;
"&="	return ANDEQ;
"|="	return IOREQ;
"^="	return XOREQ;
"%="	return MODEQ;
"<<="	return SHLEQ;
">>="	return SHREQ;
">>>="	return ZFSHRE;
"||"	return OROR;
"&&"	return ANDAND;



{letter}{alphanum}*				return screen();

0{ODigit}+[lL]?					return __INTEGER_CONSTANT ;
{DDigit}+[lL]?					return __INTEGER_CONSTANT ;
0[xX]{HDigit}+[lL]? 			return __INTEGER_CONSTANT ;

{DDigit}+\.{DDigit}*[fFdD]?((e|E)("+"|"-"){DDigit}+)?[fFdD]?	return __FLOAT_CONSTANT;
								

\'([^'\\\n]|{esc})*\'			return __CHARACTER_CONSTANT;

\"([^"\\\n]|{esc})*\"			return __STRING;


{blank}+						;
\n								++yylineno;

"/*"							BEGIN(COMMENT1);

<COMMENT1>"*/"					BEGIN(0);
<COMMENT1>[^*\n]+				;		 	/* breaks comments into lines    */
<COMMENT1>\n					++yylineno;	/* so they won't overflow buffer */
<COMMENT1>"*"					;		 	/* so they won't overflow buffer */

"//"	                      	BEGIN(COMMENT2);
<COMMENT2>\n					BEGIN(0), ++yylineno;  

{other}							return yytext[0];         


%%



/*
 * reserved word screener
*/

static const struct 
{
  	char *name;
  	int yylex;
} keywords[] = 
{
      	{"abstract",		ABSTRACT}, 
      	{"boolean",			BOOLEAN}, 
		{"break", 			BREAK},    
      	{"byte",			BYTE}, 
      	{"byvalue",			BYVALUE}, 
  		{"case",			CASE},   
      	{"cast",			CAST}, 
      	{"catch",			CATCH}, 
		{"char",			CHAR}, 
      	{"class",			CLASS}, 
		{"const",			CONST},
		{"continue",		CONTINUE},
		{"default",			DEFAULT},
		{"do",				DO},
      	{"double",			DOUBLE},
      	{"else",			ELSE},
       	{"extends",			EXTENDS},  
      	{"false",			FALSE}, 
      	{"final",			FINAL}, 
      	{"finally",			FINALLY}, 
      	{"float",			FLOAT},
      	{"for",				FOR},
      	{"future",			FUTURE}, 
      	{"generic",			GENERIC}, 
     	{"goto",			GOTO},
      	{"if",				IF},   
      	{"implements",		IMPLEMENTS}, 
      	{"import",			IMPORT}, 
      	{"inner",			INNER}, 
      	{"instanceof",		INSTANCEOF}, 
      	{"int",				INT},    
       	{"interface",		INTERFACE}, 
      	{"long",			LONG}, 
      	{"native",			NATIVE}, 
      	{"new",				NEW}, 
       	{"null",			__NULL__}, 
      	{"operator",		OPERATOR}, 
      	{"outer",			OUTER}, 
      	{"package",			PACKAGE}, 
       	{"private",			PRIVATE}, 
      	{"protected",		PROTECTED}, 
      	{"public",			PUBLIC}, 
      	{"rest",			REST}, 
      	{"return",			RETURN},
      	{"short",			SHORT},
       	{"static",			STATIC},
      	{"super",			SUPER},
      	{"switch",			SWITCH},    
      	{"synchronized",	SYNCHRONIZED}, 
       	{"this",			THIS}, 
      	{"threadsafe",		THREADSAFE}, 
      	{"throw",			THROW}, 
      	{"throws",			THROWS}, 
      	{"transient",		TRANSIENT}, 
       	{"true",			TRUE}, 
      	{"try",				TRY}, 
      	{"void",			VOID},
      	{"volatile",		VOLATILE}, 
      	{"while",			WHILE}
};

int screen(void)
{
    register int mid;
    int cc, hi, lo;

    lo = 0;
    hi = DIM(keywords) - 1;

    while (lo <= hi) 
    {
    	mid = (lo + hi) / 2;

    	if ((cc = _stricmp(yytext, keywords[mid].name)) == 0)
      		return keywords[mid].yylex;
    	if (cc < 0)
      		hi = mid - 1;
    	else
      		lo = mid + 1;
    }
    return __IDENTIFIER;
}





/*
 * error reporting code for PCYACC
 *
*/

#include <stdio.h>
#include <ctype.h>

#define WIDTH	        80			 /* width of stderr device */
#define YYERRCODE	256


FILE *yyerrfile = stderr;		  /* file to write error report to */
char  yyerrsrc[64] = "";		        /* current input file name */


/* errprefix - print where and what error message on yyerrfile 
*/
static void errprefix(char *msg)
{ 
    int punct = 0;

    fprintf(yyerrfile, "[error %d] ", error_count+1);
 
    if (yyerrsrc[0] != '\0')                        /* any input file name? */
    {		                
    	fprintf(yyerrfile, "file '%s'", yyerrsrc);
    	punct = 1;
    }

    if (yylineno >= 0)                         	      /* valid line number? */
    {
    	if (punct)
      		fprintf(yyerrfile, ", ");

    	fprintf(yyerrfile, "line %d", yylineno);
    	punct = 1;
    }

    if (yytext != NULL && *yytext != '\0') 	             /* real token? */
    {
    	if (punct)
      		fprintf(yyerrfile, " ");

    	fprintf(yyerrfile, "near \"%s\"", yytext);
        punct = 1;
    }

    if (punct)
    	fprintf(yyerrfile, ": ");

    fprintf(yyerrfile, "%s\n", msg);
}


/* yyerror - improved error reporting 
*/
void yyerror(char *s, char *t)
{ 
    static const char expecting[] = "expecting: ";
    static int list = 0;
    static int column = 0;

    if (s != NULL) 
    {
    	if (column != 0)
      		fputc('\n', yyerrfile);
    	errprefix(s);
    	if (t == NULL)
      		column = 0;
    	else
      		column = fprintf(yyerrfile, "actual: %s", t);
    	list = 0;
    } 
    else if (t != NULL) 
    {
    	if (list == 0) 
   	{
      	    if (column + strlen(t) + sizeof(expecting) + 1 < WIDTH - 2)
        	column += fprintf(yyerrfile, "  %s%s", expecting, t);
      	    else
        	column = fprintf(yyerrfile, "\n%s%s", expecting, t) - 1;
    	} 
	else 
	{
            if (column + strlen(t) < WIDTH - 2)
        	column += fprintf(yyerrfile, ", %s", t);
            else
        	column = fprintf(yyerrfile, ",\n    %s", t) - 1;
        }
        ++list;
    } 
    else 
    {
    	fprintf(yyerrfile, "\n");
    	column = list = 0;
    }
}


/* yydisplay - returns pointer to printable form for token type in 'ch' 
*/
char *  yydisplay(int ch)
{ 

    static char *tok[] = 
    {
		"ABSTRACT", 
		"BOOLEAN", 
		"BREAK",    
		"BYTE", 
		"BYVALUE", 
		"CASE",   
		"CAST", 
		"CATCH", 
		"CHAR", 
		"CLASS", 
		"CONST",
		"CONTINUE",
		"DEFAULT",
		"DO",
		"DOUBLE",
		"ELSE",
		"EXTERN",  
		"FALSE", 
		"FINAL", 
		"FINALLY", 
		"FLOAT",
		"FOR",
		"FUTURE", 
		"GENERIC", 
		"gGOTO",
		"IF",   
		"IMPLEMENTS", 
		"IMPORT", 
		"INNER", 
		"INSTANCEOF", 
		"INT",    
		"INTERFACE", 
		"LONG", 
		"NATIVE", 
		"NEW", 
		"NULL", 
		"OPERATOR", 
		"OUTER", 
		"PACKAGE", 
		"PRIVATE", 
		"PROTECTED", 
		"PUBLIC", 
		"REST", 
		"RETURN",
		"SHORT",
		"STATIC",
		"SUPER",
		"SWITCH",    
		"SYNCHRONIZED", 
		"THIS", 
		"THREADSAFE", 
		"THROW", 
		"THROWS", 
		"TRANSIENT", 
		"TRUE", 
		"TRY", 
		"VOID",
		"VOLATILE", 
		"WHILE",
	"__CHARACTER_CONSTANT",
	"__FLOAT_CONSTANT",
	"__INTEGER_CONSTANT",
	"__STRING",
	"__IDENTIFIER",  
	"ADDADD",
	"SUBSUB",
	"EQU",
	"LEQ",
	"GEQ",
	"NEQ",
	"SHL",
	"SHR",
	"ZFSHR",
	"ADDEQ",
	"SUBEQ",
	"MULEQ",
	"DIVEQ",
	"ANDEQ",
	"IOREQ",
	"XOREQ",
	"MODEQ",
	"SHLEQ",
	"SHREQ",
	"ZFSHRE",
	"OROR",
	"ANDAND",
	
		0
    };
    static char buf[16];

    switch (ch) 
    {
    	case         0: return ("[end of file]");
    	case YYERRCODE: return ("[error]");
    	case      '\a': return ("'\\a'");
    	case      '\b': return ("'\\b'");
    	case      '\f': return ("'\\f'");
    	case      '\n': return ("'\\n'");
    	case      '\r': return ("'\\r'");
    	case      '\t': return ("'\\t'");
    	case      '\v': return ("'\\v'");
    }
    if (ch > YYERRCODE && ch < YYERRCODE + DIM(tok))
    	return (tok[ch-(YYERRCODE + 1)]);		/* is %token */

    if (isascii(ch) && isprint(ch))
    	sprintf(buf, "'%c'", ch);			/* printable */
    else
    	sprintf(buf, "char %d", ch);		        /* unprintable */

    return(buf);
}
