
/*******************************************************************
*                                                                  *
*  lex.l: scanner description for Simple Calculator                *
*                                                                  *
*  By Y. Jenny Luo                                                 *
*                                                                  *
*  PCYACC (R) is a software product of ABRAXAS SOFTWARE INC.       *
*  Copyright (C) 1986-1996 by ABRAXAS SOFTWARE INC.                *
*                                                                  *
*******************************************************************/

%{

/*******************************************************************
*                                                                  *
*  lex.c: lexical analyzer for Simple Calculator                   *
*                                                                  *
*  By Y. Jenny Luo                                                 *
*                                                                  *
*  PCYACC (R) is a software product of ABRAXAS SOFTWARE INC.       *
*  Copyright (C) 1986-1996 by ABRAXAS SOFTWARE INC.                *
*                                                                  *
*******************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "public.h"             /* define global information */
#include "dcalc.h"               /* header file generated by PCYACC */


#define LEX_DEBUG       1
#define YYERRCODE       256
#define LINEWIDTH       80


unsigned char *yytext;   /* current token's text */
extern FILE *yyin;              /* pointer to input stream defined
				   in main.c */
extern char yyerrsrc[];         /* input filename defined in main.c */
extern int yyerrcnt;            /* count of error, defined in err_skel.c */

FILE *yyerrfile=stderr;

char* yydisplay(int tokenvalue); /* display the token based on tokenvalue */
void yyerror(char *s, char *t); /* improved error reporting */
static void errprefix(char *msg); /* print where and what error
				     message on yyerrfile */

int yylineno = -1;              /* current line number */
int headerflag = 0;             /* set to indicate the next header format
				   will be treated as comment. */

%}

digit           [0-9]
blank           [ \t]
sign            [+-/*]
other           .

%%

{digit}+        {
		  sscanf(yytext, "%lf", &yylval.db);
		  return NUMBER;
		}

{digit}+\.{digit}*	{
				  sscanf(yytext, "%lf", &yylval.db);
				  return NUMBER;
				}

\.{digit}+		{
				  sscanf(yytext, "%lf", &yylval.db);
				  return NUMBER;
				}


sign            return *yytext;    

{blank}+                ;

{other}                 return yytext[0];

%%

/*********************************************************************
*  FUNCTION: yydisplay                                               *
*  PARAMETER: int                                                    *
*  RETURN: int                                                       *
*  DESCRIPTION: display the token based on the input tokenvalue      *
*********************************************************************/
char* yydisplay(int tokenvalue)
{
  static char buf[TEXT];

  if (tokenvalue==NUMBER)
    return "NUMBER";

  switch (tokenvalue)
  {
    case '\a':  return ("[\\a]");
    case '\b':  return ("[\\b]");
    case '\v':  return ("[\\v]");
    case '\f':  return ("[\\f]");
    case '\r':  return ("[\\r]");
    case '\t':  return ("[\\t]");
    case '\n':  return ("[\\n]");
  }

  if (tokenvalue == 0)
  {
    return ("[end of file]");
  }
  else if (tokenvalue == YYERRCODE)
  {
    return ("[error]");
  }

  if (isprint(tokenvalue))
  {
    sprintf(buf, "%c", tokenvalue);
    return buf;
  }
  else
  {
    sprintf(buf, "[char %d]", tokenvalue);
    return buf;
  }

}

/************************************************************************
*  FUNCTION:    errprefix                                               *
*  PARAMETER:   char *                                                  *
*  RETURN:      none                                                    *
*  DESCRIPTION: print where and what error message on error file        *
************************************************************************/
static void errprefix(char *msg)
{
  int punct=0;

  fprintf(yyerrfile, "[error %d] ", yyerrcnt+1);
  if (yyerrsrc[0]!='\0')                        /* any input file name? */
  {
    fprintf(yyerrfile, "file '%s'", yyerrsrc);
    punct=1;
  } /* end if */

  if (yylineno>=0)                                    /* valid line number? */
  {
    if (punct)
    {
      fprintf(yyerrfile, ", ");
    } /*end if */
    fprintf(yyerrfile, "line %d", yylineno);
    punct=1;
  } /* end if */

  if (yytext==0&&*yytext!='\0')                      /* real token? */
  {
    if (punct)
    {
      fprintf(yyerrfile, " ");
    } /* end if */
    fprintf(yyerrfile, "near \"%s\"", yytext);
    punct=1;
  } /* end if */

  if (punct)
  {
    fprintf(yyerrfile, ": ");
  } /* end if */
  fprintf(yyerrfile, "%s\n", msg);
} /* end errprefix */

/************************************************************************
*  FUNCTION:    yyerror                                                 *
*  PARAMETER:   char *,  char *                                         *
*  RETURN:      none                                                    *
*  DESCRIPTION: for improved error reporting                            *
************************************************************************/
void yyerror(char *s, char *t)
{
  static const char *expecting="expecting: ";
  static int list=0;
  static int column=0;

  if (s!=0)
  {
    if (column!=0)
    {
      fputc('\n', yyerrfile);
    } /*end if */
    errprefix(s);
    if (t==0)
    {
      column=0;
    }
    else
    {
      column=fprintf(yyerrfile,"actual: %s", t);
    } /* end if */
    list=0;
  }
  else
  {
    if (t!=0)
    {
      if (list == 0)
      {
	if (column+strlen(t)+sizeof(expecting)+1<LINEWIDTH-2)
	{
	  column+=fprintf(yyerrfile, "  %s%s", expecting, t);
	}
	else
	{
	  column=fprintf(yyerrfile, "\n%s%s", expecting, t)-1;
	} /* end if */
      }
      else
      {
	if (column+strlen(t)<LINEWIDTH-2)
	{
	  column+=fprintf(yyerrfile, ", %s", t);
	}
	else
	{
	  column=fprintf(yyerrfile, ",\n    %s", t)-1;
	} /* end if */
      } /* end if */
      ++list;
    }
    else
    {
      fprintf(yyerrfile,"\n");
      column=0;
      list=0;
    } /* end if */
  } /* end if */
} /* end yyerror */
